import os
import shutil


def _add_cls(install, app):
    if install.cls:
        os.makedirs('./app/extension/' + app, exist_ok=True)
        for item in install.cls:
            with open(f'./app/extension/{app}/{item[0]}.py', 'a') as file:
                f = ''
                f += f'from {item[1]} import {item[2]}\n\n'
                f += f'class {item[3]}({item[2]}):\n'
                f += '    pass\n'
                file.write(f)


def _add_imprt(install, app):
    if install.imprt:
        os.makedirs('./app/extension/' + app, exist_ok=True)
        for item in install.imprt:
            with open(f'./app/extension/{app}/{item[0]}.py', 'a') as file:
                f = f'from {item[1]} import {item[2]}\n'
                file.write(f)


def _add_route(install, app):
    if install.route:
        os.makedirs('./app/extension/' + app, exist_ok=True)
        with open(f'./app/extension/{app}/route.py', 'w') as file:
            f = 'import ' + install.route + '\n'
            file.write(f)

        with open('./routes/index.py', 'r') as file:
            all = file.read()
            routes = 'import ' + f'app.extension.{app.replace("/", ".")}.route' + '\n'
            all = routes + all
            with open('./routes/index.py', 'w') as b:
                b.write(all)


def _remove_route(install, app):
    if install.route:
        with open('./routes/index.py', 'r') as file:
            all = file.read()
            routes = 'import ' + f'app.extension.{app.replace("/", ".")}.route' + '\n'
            all = all.replace(routes, '')
            with open('./routes/index.py', 'w') as b:
                b.write(all)


def _add_admin_templates(install, app):
    name = app.replace("/", "_")
    shutil.copytree(install.admin_templates, './frontend/admin/src/extension/' + app)
    with open('./frontend/admin/src/router/index.js', 'r') as file:
        all = file.read()
        all = "import " + name + "_routes from '@/extension/" + app + "/route.js'\n" + all
        all = all.replace('/* {{place new route}} */',
                          '/* {{place new route}} */\n...' + name + '_routes,')

        with open('./frontend/admin/src/router/index.js', 'w') as b:
            b.write(all)


def _remove_admin_templates(app):
    name = app.replace("/", "_")
    shutil.rmtree('./frontend/admin/src/extension/' + app)
    with open('./frontend/admin/src/router/index.js', 'r') as file:
        all = file.read()
        all = all.replace("import " + name + "_routes from '@/extension/" + app + "/route.js'\n", '')
        all = all.replace('...' + name + '_routes,', '')

        with open('./frontend/admin/src/router/index.js', 'w') as b:
            b.write(all)


def _add_index_templates(install, app):
    name = app.replace("/", "_")
    shutil.copytree(install.index_templates, './frontend/index/src/extension/' + app)
    with open('./frontend/index/src/router/index.js', 'r') as file:
        all = file.read()
        all = "import " + name + "_routes from '@/extension/" + app + "/route.js'\n" + all

        all = all.replace('/* {{place new route}} */',
                          '/* {{place new route}} */\n...' + name + '_routes,')

        with open('./frontend/index/src/router/index.js', 'w') as b:
            b.write(all)


def _remove_index_templates(app):
    name = app.replace("/", "_")
    shutil.rmtree('./frontend/index/src/extension/' + app)
    with open('./frontend/index/src/router/index.js', 'r') as file:
        all = file.read()
        all = all.replace("import " + name + "_routes from '@/extension/" + app + "/route.js'\n", '')
        all = all.replace('...' + name + '_routes,', '')

        with open('./frontend/index/src/router/index.js', 'w') as b:
            b.write(all)


def _add_component_templates(install, app):
    shutil.copytree(install.component_templates, './frontend/index/src/components/' + app)
    shutil.copytree(install.component_templates, './frontend/admin/src/components/' + app)


def _remove_component_templates(app):
    shutil.rmtree('./frontend/index/src/components/' + app)
    shutil.rmtree('./frontend/admin/src/components/' + app)


def _add_config(install):
    if install.config:
        with open(f'./config/{install.config[0]}.py', 'w') as b:
            b.write(open(install.config[1], 'r').read())


def _remove_config(install):
    if install.config:
        os.remove(f'./config/{install.config[0]}.py')



import shutil

from renus.commands.extension.build import _remove_route, _remove_admin_templates, _remove_index_templates, \
    _remove_config, _remove_component_templates
from renus.commands.help import bc


def remove_all(installed: dict):
    items = list(installed.copy().keys())
    items.reverse()
    for app in items:
        remove(app, installed)


def remove(app, installed):
    print(f'   remove {app}')
    del installed[app]
    try:
        install = __import__(f"extension.{app.replace('/', '.')}.install", fromlist=[''])

    except Exception as exc:
        print('   ' + bc.WARNING + str(exc) + bc.ENDC)
        return

    if hasattr(install, 'setup_remove'):
        print(f'   setup remove {app}...')
        try:
            install.setup_remove()
        except Exception as exc:
            print('   ' + bc.FAIL + str(exc) + bc.ENDC)

    if hasattr(install, 'route'):
        print(f'   remove {app} route...')
        _remove_route(install, app)

    if hasattr(install, 'admin_templates'):
        print(f'   remove {app} admin_templates...')
        try:
            _remove_admin_templates(app)
        except Exception as exc:
            print('   ' + bc.FAIL + str(exc) + bc.ENDC)

    if hasattr(install, 'index_templates'):
        print(f'   remove {app} index_templates...')
        try:
            _remove_index_templates(app)
        except Exception as exc:
            print('   ' + bc.FAIL + str(exc) + bc.ENDC)

    if hasattr(install, 'component_templates'):
        print(f'   remove {app} component_templates...')
        try:
            _remove_component_templates(app)
        except Exception as exc:
            print('   ' + bc.FAIL + str(exc) + bc.ENDC)

    if hasattr(install, 'config'):
        print(f'   remove {app} config...')
        try:
            _remove_config(install)
        except Exception as exc:
            print('   ' + bc.FAIL + str(exc) + bc.ENDC)

    print(f'   remove {app} folder...')
    try:
        shutil.rmtree('./app/extension/' + app)
        shutil.rmtree('./extension/' + app)
    except Exception as exc:
        print('   ' + bc.FAIL + str(exc) + bc.ENDC)



import base64
import os
import random


class Cryptor:
    """A cryptographic utility for encrypting and decrypting text using a matrix-based transformation.

    The Cryptor uses a combination of shuffling, mixing, and matrix operations to obscure the
    original text. It supports configurable matrix sizes for the transformation process.
    """

    def __init__(self):
        """Creates a new Cryptor instance with default matrix size (32)."""
        self.matrix = 32

    def encrypt(self, data: bytes, key: str) -> bytes:
        """Encrypts raw bytes using the provided key.

        Args:
            data: The bytes to encrypt
            key: The encryption key

        Returns:
            The encrypted bytes

        Raises:
            ValueError: If encryption fails
        """
        matrix_size = self.matrix
        key_bytes = self._generate_password(matrix_size, key.encode())
        data_size = len(data).to_bytes(4, byteorder='big')
        random_prefix = self._get_random_bytes(6)
        seed_random = sum(b for b in random_prefix)

        padded_text = bytearray()
        padded_text.extend(data_size)
        padded_text.extend(random_prefix)
        padded_text.extend(data)

        seed_sum = sum(b for b in key_bytes)
        self._shuffle(padded_text, seed_sum + seed_random, 5)

        # Split into matrix chunks
        matrix = [padded_text[i:i + matrix_size] for i in range(0, len(padded_text), matrix_size)]

        for i in range(len(matrix)):
            seed = matrix[i + 1][0] if i + 1 < len(matrix) else key_bytes[0]
            self._shuffle(matrix[i], seed + seed_random, 2)

        # Flatten the matrix back to bytes
        padded_text = bytearray().join(matrix)
        self._mix(matrix_size, padded_text, key_bytes)

        padded_text.extend(seed_random.to_bytes(2, byteorder='big'))
        return bytes(padded_text)

    def encrypt_text(self, text: str, key: str) -> str:
        """Encrypts text using the provided key and returns a URL-safe base64 string.

        Args:
            text: The plaintext to encrypt
            key: The encryption key

        Returns:
            URL-safe base64 encoded encrypted string without padding
        """
        encrypted = self.encrypt(text.encode(), key)
        return base64.urlsafe_b64encode(encrypted).decode().rstrip('=')

    def decrypt(self, encoded: bytes, key: str) -> bytes:
        """Decrypts bytes using the provided key.

        Args:
            encoded: The encrypted bytes to decrypt
            key: The decryption key

        Returns:
            The decrypted bytes

        Raises:
            ValueError: If decryption fails
        """
        if len(encoded) < 6:
            raise ValueError("Invalid Token Matrix Length")

        seed_random = int.from_bytes(encoded[-2:], byteorder='big')
        decoded = bytearray(encoded[:-2])
        matrix_size = self.matrix

        key_bytes = self._generate_password(matrix_size, key.encode())
        self._unmix(matrix_size, decoded, key_bytes)

        # Split into matrix chunks
        matrix = [decoded[i:i + matrix_size] for i in range(0, len(decoded), matrix_size)]

        for i in reversed(range(len(matrix))):
            seed = matrix[i + 1][0] if i + 1 < len(matrix) else key_bytes[0]
            self._unshuffle(matrix[i], seed + seed_random, 2)

        # Flatten the matrix back to bytes
        decoded = bytearray().join(matrix)
        seed_sum = sum(b for b in key_bytes)
        self._unshuffle(decoded, seed_sum + seed_random, 5)

        data_size = int.from_bytes(decoded[:4], byteorder='big')
        if len(decoded) < data_size + 10:
            raise ValueError("Invalid Token Matrix Length")

        return bytes(decoded[10:data_size + 10])

    def decrypt_text(self, encoded: str, key: str) -> str:
        """Decrypts a URL-safe base64 encoded string using the provided key.

        Args:
            encoded: URL-safe base64 encoded string to decrypt
            key: The decryption key

        Returns:
            The decrypted string

        Raises:
            ValueError: If decryption fails
        """
        # Add padding if needed
        padding = len(encoded) % 4
        if padding != 0:
            encoded += '=' * (4 - padding)

        data = base64.urlsafe_b64decode(encoded)
        return self.decrypt(data, key).decode()

    def set_matrix(self, size: int):
        """Sets the matrix size used for cryptographic operations.

        The matrix size determines how data is chunked and processed during encryption/decryption.
        Must be a positive non-zero value.
        """
        if size > 0:
            self.matrix = size

    # Utility methods
    def _generate_password(self, matrix: int, password: bytes) -> bytes:
        """Generates a key of specified length by repeating the password."""
        if not password:
            return bytes(matrix)

        password_len = len(password)
        repeats = matrix // password_len
        remainder = matrix % password_len

        result = bytearray()
        for _ in range(repeats):
            result.extend(password)
        if remainder > 0:
            result.extend(password[:remainder])

        return bytes(result)

    def _shuffle(self, data: bytearray, seed: int, step: int):
        """Shuffles the data using the given seed."""
        random.seed(seed)
        length = len(data)

        for i in reversed(range(1, length, step)):
            j = random.randint(0, i)
            data[i], data[j] = data[j], data[i]

    def _unshuffle(self, data: bytearray, seed: int, step: int):
        """Reverses the shuffling operation."""
        random.seed(seed)
        length = len(data)
        swaps = []

        for i in reversed(range(1, length, step)):
            j = random.randint(0, i)
            swaps.append((i, j))

        for i, j in reversed(swaps):
            data[i], data[j] = data[j], data[i]

    def _mix(self, block_size: int, buf: bytearray, key: bytes):
        """Mixes the data blocks with XOR operations."""
        prev_block = key

        for i in range(0, len(buf), block_size):
            block = buf[i:i + block_size]
            for j in range(len(block)):
                block[j] ^= prev_block[j]
            prev_block = block

    def _unmix(self, block_size: int, buf: bytearray, key: bytes):
        """Reverses the mixing operation."""
        chunks = [buf[i:i + block_size] for i in range(0, len(buf), block_size)]

        for i in reversed(range(len(chunks))):
            if i == 0:
                for j in range(len(chunks[i])):
                    chunks[i][j] ^= key[j]
            else:
                for j in range(len(chunks[i])):
                    chunks[i][j] ^= chunks[i - 1][j]

    def _get_random_bytes(self, length: int) -> bytes:
        """Generates cryptographically secure random bytes."""
        return os.urandom(length)



from datetime import datetime
from typing import TypeVar, Generic, Any, List, Optional, Union, Tuple, Dict

from bson import ObjectId
from pymongo import MongoClient

from renus.core.config import Config
from renus.core.cprint import Cprint


class ReModel:
    def __init__(self, doc):
        for k, v in doc.items():
            setattr(self, k, v)

    def __iter__(self):
        return iter(self.__dict__.items())

    def __getitem__(self, item):
        return self.__dict__[item]


M = TypeVar('T')


class ModelBase(Generic[M]):
    client = MongoClient(Config('database').get('host', '127.0.0.1'),
                         Config('database').get('port', 27017),
                         username=Config('database').get('username', None),
                         password=Config('database').get('password', None))

    _database_name = Config('database').get('name', 'renus')
    collection_name = None
    metro = None
    storage = None
    hidden_fields = []
    document_model = dict
    add_time_fields = True

    def __init__(self) -> None:
        if hasattr(self, '_collection_name'):
            raise RuntimeError('_collection_name has expired. Use collection_name')

        self._steps = None
        self._where = None
        self._distinct = None
        self._limit = None
        self._skip = None
        self._sort = None
        self._select = None
        self.visible_fields = []

    def __iter__(self):
        return iter(self.get())

    def set_database(self, name: str):
        self._database_name = name
        return self

    def set_collection(self, name: str):
        self.collection_name = name
        return self

    @property
    def db(self):
        return self.client[self._database_name]

    def collection(self, name: str = None):
        if name is None:
            name = self.collection_name

        return self.db[name]

    def reset(self):
        self._steps = None
        self._where = None
        self._distinct = None
        self._limit = None
        self._skip = None
        self._sort = None
        self._select = None
        self.visible_fields = []
        return self

    def aggregate(self, pipeline: Any, session: Any = None):
        return self.__police(self.collection().aggregate(pipeline, session))

    def where(self, where: dict):
        if self._steps is not None:
            self._steps.append({
                "$match": where
            })
        else:
            if self._where is None:
                self._where = {}
            for k, v in where.items():
                if k.startswith('$') and k in self._where:
                    if type(v) is list:
                        self._where[k] += v
                    else:
                        self._where[k] = {**self._where[k], **v}
                else:
                    self._where[k] = v

        return self

    def count(self, skip=False, limit=False):
        k = {}
        if self._where is not None:
            k['filter'] = self._where
        else:
            k['filter'] = {}
        if skip and self._skip is not None:
            k['skip'] = self._skip
        if limit and self._limit is not None:
            k['limit'] = self._limit

        return self.collection().count_documents(**k)

    def distinct(self, key: str):
        self._distinct = key
        return self

    def limit(self, count: int):
        if self._steps is not None:
            self._steps.append({"$limit": count})
        else:
            self._limit = count
        return self

    def skip(self, count: int):
        if self._steps is not None:
            self._steps.append({"$skip": count})
        else:
            self._skip = count
        return self

    def group(self, item: dict):
        if self._steps is None:
            raise RuntimeError("group work with steps() architect")
        self._steps.append({"$group": item})
        return self

    def sort(self, key_or_list: Union[str, List[Tuple]], asc: bool = True):
        if type(key_or_list) is dict:
            s = key_or_list
        elif type(key_or_list) is list:
            s = []
            for f, a in key_or_list:
                s.append((f, 1 if a else -1))
        else:
            s = [(key_or_list, 1 if asc else -1)]
        if self._steps is not None:
            if type(s) is list:
                s = {i[0]: i[1] for i in s}
            self._steps.append({"$sort": s})
        else:
            self._sort = s
        return self

    def select(self, *select: [str, Dict]):
        s = {}
        if len(select) == 1 and type(select[0]) is dict:
            s = select[0]
        else:
            for key in select:
                if type(key) is dict:
                    s = key
                else:
                    s[key] = 1

        if self._steps is not None:
            self._steps.append({'$project': s})
            return self

        self._select = s
        return self

    def with_relation(self, collection, local_field, forigen_field, to, single=False, pipeline=None):
        if pipeline is None:
            pipeline = []
        if self._steps is not None:
            self._steps.append({
                "$lookup": {
                    "from": collection,
                    "localField": local_field,
                    "foreignField": forigen_field,
                    "as": to,
                    "pipeline": pipeline
                }
            })
            if single:
                self._steps.append({'$addFields': {f'{to}': {'$arrayElemAt': [f"${to}", 0]}}})
        else:
            c = Cprint()
            c.print(c.red('use steps mode for with_relation.'))
        return self

    def steps(self):
        self._steps = []
        return self

    def get(self, police: bool = True, session=None) -> List[M]:
        if self._steps is not None:
            return self.aggregate(self._steps, session)

        where, select = self._base_gate(police)

        find = self.collection().find(where, select, session=session)

        if self._sort is not None:
            find = find.sort(self._sort)
        if self._skip is not None:
            find = find.skip(self._skip)
        if self._limit is not None:
            find = find.limit(self._limit)
        if self._distinct is not None:
            return find.distinct(self._distinct)

        return self.__police(find) if police else list(find)

    def first(self, police: bool = True, session=None) -> Optional[M]:
        self.limit(1)
        res = self.get(police, session)
        self._limit = None
        if len(res) == 1:
            return res[0]
        return None

    def create(self, document: dict, session=None) -> dict:
        if self.add_time_fields and "updated_at" not in document:
            document["updated_at"] = datetime.utcnow()

        if self.add_time_fields and "created_at" not in document:
            document["created_at"] = datetime.utcnow()

        id = self.collection().insert_one(document, session=session).inserted_id
        document['_id'] = id
        self.boot_event('create', {}, document, session)
        self._attach_file(document, session)
        return document

    def create_many(self, documents: List, session=None) -> List[ObjectId]:
        if self.add_time_fields:
            for document in documents:
                if "updated_at" not in document:
                    document["updated_at"] = datetime.utcnow()

                if "created_at" not in document:
                    document["created_at"] = datetime.utcnow()

        ids = self.collection().insert_many(documents, session=session).inserted_ids
        self.boot_event('create_many', {}, ids, session)
        self._attach_file({'_id': 'create_many', 'documents': documents}, session)
        return ids

    def update(self, new: dict, upsert=False, get_old=False, session=None) -> bool:
        where = self.__ud_gate('update')
        if self.add_time_fields:
            new["updated_at"] = datetime.utcnow()
        d = {"$set": new}
        if self.add_time_fields and "created_at" not in new:
            d["$setOnInsert"] = {'created_at': datetime.utcnow()}
        old = self.collection().find_one_and_update(where, d, upsert=upsert, session=session)
        self.boot_event('update', old, new, session)
        if old is None:
            new = {'_id': 'upsert', 'doc': new}
        else:
            new['_id'] = old['_id']
        self._attach_file(new, session)
        return old if get_old else True

    def update_opt(self, new: dict, upsert=False, get_old=False, session=None) -> bool:
        where = self.__ud_gate('update')
        if '$set' not in new:
            new['$set'] = {}
        if '$setOnInsert' not in new:
            new['$setOnInsert'] = {}
        if self.add_time_fields:
            new['$set']["updated_at"] = datetime.utcnow()
            new['$setOnInsert']['created_at'] = datetime.utcnow()
        old = self.collection().find_one_and_update(where, new, upsert=upsert, session=session)
        self.boot_event('update', old, {k[1:]: v for k, v in new.items()}, session)
        if old is None:
            new = {'_id': 'upsert', 'doc': new}
        else:
            new['_id'] = old['_id']
        self._attach_file(new, session)
        return old if get_old else True

    def update_opt_many(self, new: dict, upsert=False, get_old=False, session=None) -> bool:
        where = self.__ud_gate('update')
        if '$set' not in new:
            new['$set'] = {}
        if '$setOnInsert' not in new:
            new['$setOnInsert'] = {}
        if self.add_time_fields:
            new['$set']["updated_at"] = datetime.utcnow()
            new['$setOnInsert']['created_at'] = datetime.utcnow()
        old = self.collection().update_many(where, new, upsert=upsert, session=session).raw_result
        self.boot_event('update_many', old, {k[1:]: v for k, v in new.items()}, session)
        self._attach_file({'_id': 'update_opt_many', 'documents': new}, session)
        return old if get_old else True

    def update_many(self, new: dict, session=None) -> bool:
        where = self.__ud_gate('update')
        if self.add_time_fields:
            new["updated_at"] = datetime.utcnow()
        old = self.collection().update_many(where, {"$set": new}, session=session).raw_result
        self.boot_event('update_many', old, str(new), session)
        self._attach_file({'_id': 'update_many', 'documents': new}, session)
        return True

    def delete(self, all: bool = False, session=None) -> bool:
        where = self.__ud_gate('delete')
        if all is False:
            old = self.collection().find_one_and_delete(where, session=session)
            if self.metro is not None:
                self._handle_metro('delete', old, session=session)
            self.boot_event('delete', old, {}, session)
        else:
            if self.metro is not None:
                self._handle_metro('delete', session=session)
            old = self.collection().delete_many(where, session=session)

            self.boot_event('delete_many', {'deleted_count': old.deleted_count, 'where': str(where)}, {}, session)

        return old

    def make_visible(self, fields: List):
        self.visible_fields = fields
        return self

    @staticmethod
    def boot_event(typ: str, old, new, session):
        pass

    @staticmethod
    def convert_id(id):
        if type(id) is str:
            try:
                return ObjectId(id)
            except Exception:
                return id
        return id

    def _base_gate(self, police) -> list:
        where = self._where
        select = self._select

        if select is None and police == False:
            raise RuntimeError("when police is off, for security reason select fields. ex: .select('_id','name')")

        if where is not None:
            if '_id' in where:
                where['_id'] = self.convert_id(where['_id'])

        return [where, select]

    @staticmethod
    def cast(document: dict):
        return document

    def __cleaner(self, document: dict):
        for field in self.hidden_fields:
            if field in document and field not in self.visible_fields:
                del document[field]

        return self.document_model(self.cast(document))

    def __police(self, documents: Any):
        if documents is None:
            return None
        res = []
        if type(documents) is dict:
            return self.__cleaner(documents)
        else:
            for document in documents:
                if type(document) is dict:
                    res.append(self.__cleaner(document))
        return res

    def __ud_gate(self, type: str):
        where = self._where
        if where is None:
            raise RuntimeError(f"for {type} use where. ex: .where({'name':'test'})")
        if '_id' in where:
            where['_id'] = self.convert_id(where['_id'])

        return where

    def _handle_metro(self, typ, obj=None, session=None):
        if typ == 'delete':
            if obj is None:
                where = self.__ud_gate('delete')
                all = self.collection().find(where, session=session)
            else:
                all = [obj]

            for item in all:
                for field, db in self.metro.items():
                    if type(db) is not list:
                        raise RuntimeError(
                            "metro format not true. ex: '_id':[{'collection':'test','field': 'test_id'}]")
                    for d in db:
                        c = d.get('collection', False)
                        if c is not False:
                            if self.storage:
                                ids = list(self.collection(c).find({
                                    d['field']: item[field]
                                }, {"_id": 1}, session=session))
                                for id_item in ids:
                                    self._detach_file(id_item, session)
                            self.collection(c).delete_many({
                                d['field']: item[field]
                            }, session=session)

                self._detach_file(item, session)

    def __get_links(self, field: str, doc: dict):
        item = doc.copy()
        p = field.split('.')
        r = None
        for i in p:
            if type(item) is list:
                res = []
                for f in item:
                    if i not in f:
                        return False
                    res.append(f[i])
                return res
            else:
                if i not in item:
                    return False
                r = item[i]
                item = item[i]
        return r

    def _attach_file(self, item, session):
        if self.storage is None or self.metro is None:
            return
        links = self._links_extractor(item)
        self.storage.reset().where({
            'path': {'$in': links}
        }).update_many({'type': self.collection_name, 'type_id': item['_id']}, session=session)

    def _detach_file(self, item, session):
        if self.storage is None or self.metro is None:
            return
        self.storage.reset().where({
            'type_id': item['_id']
        }).update_many({'type': None, 'type_id': None}, session=session)


    @property
    def database_name(self):
        return self._database_name

    def _links_extractor(self, item) -> list:
        s = []
        for field, db in self.metro.items():
            if type(db) is not list:
                raise RuntimeError(
                    "metro format not true. ex: '_id':[{'collection':'test','field': 'test_id'}]")
            for d in db:
                strg = d.get('storage', False)
                if strg is not False:
                    l = self.__get_links(strg, item)
                    if l:
                        s.extend(l)

        return s



import shutil
import zipfile
from importlib.metadata import version as md_version
from io import BytesIO

import requests

from renus.commands.extension.build import _add_route, _add_admin_templates, _add_index_templates, \
    _add_cls, _add_imprt, _add_config, _add_component_templates
from renus.commands.extension.remove import remove_all, remove
from renus.commands.help import bc


def download(app, version, installed, isUpdate=False):
    if app in installed:
        if isUpdate and installed[app] == version:
            return version
        elif not isUpdate:
            return version

    print(f'{bc.OKGREEN}downloading {app}/{version}{bc.ENDC}')

    res = requests.get(f'https://codenus.com/app/{app}/{version}')

    if res.status_code != 200:
        print(f'{bc.FAIL}download error{bc.ENDC}')
        return

    if app in installed:
        if isUpdate and installed[app] == res.headers['version']:
            print('   same version', res.headers['version'])
            return res.headers['version']
        elif not isUpdate:
            print('   currently installed version', installed[app])
            return res.headers['version']

    installed[app] = res.headers['version']
    f = zipfile.ZipFile(BytesIO(res.content))
    shutil.rmtree(f"extension/{app}", True)
    f.extractall(f"extension/{app}")

    install_app(app, installed, isUpdate)
    if isUpdate:
        print(f'{bc.OKBLUE}   updated to version {res.headers["version"]}{bc.ENDC}')
    return res.headers['version']


def rewrite(installed):
    print('   rewrite installed')
    pre_line = ''
    with open('extension/install.py', 'r') as f:
        lines = f.readlines()
        for line in lines:
            if line.find('installed') != -1:
                break
            pre_line += line
    with open('extension/install.py', 'w') as f:
        f.write(pre_line)
        f.write('installed = {\n')
        for app, version in installed.items():
            f.write(f'  "{app}":"{version}",\n')
        f.write('}')


def check_version(install, app):
    renus_requires = install.renus_requires
    renus_version = md_version('renus').split('.')
    min = renus_requires['min'].split('.')
    max = renus_requires['max'].split('.')
    re_version = 0
    min_version = 0
    max_version = 0
    for i in range(len(renus_version)):
        re_version += int(renus_version[i]) * (10000 ** (len(renus_version) - i))

    for i in range(len(min)):
        min_version += int(min[i]) * (10000 ** (len(renus_version) - i))

    for i in range(len(max)):
        max_version += int(max[i]) * (10000 ** (len(renus_version) - i))
    if re_version < min_version or re_version >= max_version:
        raise RuntimeError(f"{app}: renus version too high. {renus_version} >= {renus_requires['max']}")


def install_app(app, installed, isUpdate):
    print(f'   install {app}')
    try:
        install = __import__(f"extension.{app.replace('/', '.')}.install", fromlist=[''])

    except Exception as exc:
        print('   ' + bc.WARNING + str(exc) + bc.ENDC)
        return

    check_version(install, app)

    if hasattr(install, 'dependencies'):
        print(f'   install {app} dependencies...')
        for dep in install.dependencies:
            download(dep, '*', installed)

    if not isUpdate:
        if hasattr(install, 'cls'):
            print(f'   add {app} class...')
            try:
                _add_cls(install, app)
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)

        if hasattr(install, 'route'):
            print(f'   add {app} route...')
            _add_route(install, app)

        if hasattr(install, 'imprt'):
            print(f'   add {app} imports...')
            try:
                _add_imprt(install, app)
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)

        if hasattr(install, 'admin_templates'):
            print(f'   add {app} admin_templates...')
            try:
                if install.admin_templates:
                    _add_admin_templates(install, app)
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)

        if hasattr(install, 'index_templates'):
            print(f'   add {app} index_templates...')
            try:
                if install.index_templates:
                    _add_index_templates(install, app)
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)

        if hasattr(install, 'component_templates'):
            print(f'   add {app} component_templates...')
            try:
                if install.component_templates:
                    _add_component_templates(install, app)
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)

        if hasattr(install, 'config'):
            print(f'   add {app} config...')
            try:
                _add_config(install)
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)

        if hasattr(install, 'setup'):
            print(f'   setup {app}...')
            try:
                install.setup()
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)
    else:
        if hasattr(install, 'setup_update'):
            print(f'   setup update {app}...')
            try:
                install.setup_update()
            except Exception as exc:
                print('   ' + bc.FAIL + str(exc) + bc.ENDC)


def install_all(apps, installed, isUpdate):
    d = apps
    if isUpdate:
        d = installed
    for app, version in d.items():
        download(app, '*' if isUpdate else version, installed, isUpdate)


def run(args=None):
    print('start...')
    try:
        from extension.install import apps, installed
    except:
        apps = {}
        installed = {}

    isUpdate = False
    isRemove = False

    if args and '--update' in args:
        isUpdate = True
        args.remove('--update')
    if args and '--remove' in args:
        isRemove = True
        args.remove('--remove')
    if isUpdate and isRemove:
        raise RuntimeError("use one of '--update' or '--remove'")

    if args is None or len(args) == 0:
        if isRemove:
            remove_all(installed)
        else:
            install_all(apps, installed, isUpdate)
    else:
        s = args[0].split('==')
        version = "*"
        if len(s) > 1:
            version = s[1]
        if isRemove:
            remove(s[0], installed)
        else:
            download(s[0], version, installed, isUpdate)

    rewrite(installed)

    print(f'{bc.OKBLUE}finished{bc.ENDC}')
